#
# dcolumn/common/model_mixins.py
#

"""
Mixins used in Django models.

by: Carl J. Nobile

email: carl.nobile@gmail.com
"""
__docformat__ = "restructuredtext en"

import re
import logging
from datetime import datetime
from dateutil.tz import tzutc

from django.db import models
from django.db.models import Q
from django.contrib.auth.models import User
from django.utils.translation import ugettext_lazy as _

from dcolumn.common import ChoiceManagerImplementation

log = logging.getLogger('dcolumn.models')


#
# UserModel
#
class UserModelMixin(models.Model):
    """
    Abstract model mixin used in the model classes to provide user and
    creator fields.
    """

    updater = models.ForeignKey(
        User, verbose_name=_("Updater"), db_index=True, editable=False,
        related_name="%(app_label)s_%(class)s_updater_related",
        help_text=_("The last user to update this record."))
    creator = models.ForeignKey(
        User, verbose_name=_("Creator"), db_index=True, editable=False,
        related_name="%(app_label)s_%(class)s_creator_related",
        help_text=_("The user  who created this record."))

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        """
        Save is here to assure that save is executed throughout the MRO.
        """
        super(UserModelMixin, self).save(*args, **kwargs)

    def _updater_producer(self):
        """
        Primary use is in the admin class to supply the user's full name.
        """
        return self.updater.get_full_name()
    _updater_producer.short_description = _("Updater")

    def _creator_producer(self):
        """
        Primary use is in the admin class to supply the creator's full name.
        """
        return self.creator.get_full_name()
    _creator_producer.short_description = _("Creator")


#
# TimeModel
#
class TimeModelMixin(models.Model):
    """
    Abstract model mixin used in the model classes to supply created and
    updated fields.
    """

    created = models.DateTimeField(
        verbose_name=_("Date Created"),
        help_text=_("The date and time of creation."))
    updated = models.DateTimeField(
        verbose_name=_("Last Updated"),
        help_text=_("The date and time last updated."))

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        """
        Permit the disabling of the created and updated date times.
        """
        if not kwargs.pop(u'disable_created', False) and self.created is None:
            self.created = datetime.now(tzutc())

        if not kwargs.pop(u'disable_updated', False):
            self.updated = datetime.now(tzutc())

        log.debug("kwargs: %s, created: %s, updated: %s",
                  kwargs, self.created, self.updated)
        super(TimeModelMixin, self).save(*args, **kwargs)


#
# StatusModel
#
class StatusModelManagerMixin(models.Manager):
    """
    Manager mixin for the StatusModelMixin abstract model.
    """

    def active(self, active=True):
        """
        Return as default only active database objects.

        :Parameters:
          active : `bool`
            If `True` return only active records else if `False` return
            non-active records. If `None` return all records.
        """
        query = []

        if active is not None:
            query.append(Q(active=active))

        return self.filter(*query)


class StatusModelMixin(models.Model):
    """
    Abstract model mixin used in the model classes to supply the active field.
    """

    active = models.BooleanField(
        verbose_name=_("Active"), default=True,
        help_text=_("If checked the record is active."))

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        """
        Save is here to assure that save is executed throughout the MRO.
        """
        super(StatusModelMixin, self).save(*args, **kwargs)


#
# BaseChoiceModelManager
#
class BaseChoiceModelManager(models.Manager, ChoiceManagerImplementation):

    def get_value_by_pk(self, pk, field=None):
        value = u''

        if int(pk) != 0:
            try:
                obj = self.get(pk=pk)
                value = getattr(obj, field)
            except self.model.DoesNotExist as e:
                log.error("Access to PK %s failed, %s", pk, e)

        return value

    def get_choices(self, field, comment=True):
        choices = [(obj.pk, getattr(obj, field)) for obj in self.all()]

        if comment:
            choices.insert(
                0, (0, _("Please choose a {}".format(self.model.__name__))))

        return choices



#
# dcolumn/common/view_mixins.py
#

import logging, json
from django.http import HttpResponse

log = logging.getLogger('dcolumn.views')


class JSONResponseMixin(object):
    """
    A mixin that can be used to render a JSON response.
    """
    response_class = HttpResponse

    def render_to_response(self, context, **response_kwargs):
        """
        Returns a JSON response, transforming 'context' to make the payload.
        """
        response_kwargs['content_type'] = 'application/json'
        return self.response_class(self.convert_context_to_json(context),
                                   **response_kwargs)

    def convert_context_to_json(self, context):
        """
        Convert the context dictionary into a JSON object
        """
        return json.dumps(context)


class AjaxableResponseMixin(object):
    """
    Mixin to add AJAX support to a form.
    Must be used with an object-based FormView (e.g. CreateView)
    """
    def render_to_json_response(self, context, **response_kwargs):
        data = json.dumps(context)
        #log.debug("data: %s", data)
        response_kwargs['content_type'] = 'application/json'
        return HttpResponse(data, **response_kwargs)

    def form_invalid(self, form):
        response = super(AjaxableResponseMixin, self).form_invalid(form)

        if self.request.is_ajax():
            return self.render_to_json_response(form.errors, status=400)
        else:
            return response

    def form_valid(self, form):
        # We make sure to call the parent's form_valid() method because
        # it might do some processing (in the case of CreateView, it will
        # call form.save() for example).
        response = super(AjaxableResponseMixin, self).form_valid(form)

        if self.request.is_ajax():
            data = {'pk': self.object.pk}
            return self.render_to_json_response(data)
        else:
            return response



#
# dcolumn/dcolumns/models.py
#

"""
Dynamic Column depended models

by: Carl J. Nobile

email: carl.nobile@gmail.com
"""
__docformat__ = "restructuredtext en"

import logging
from collections import OrderedDict

from django.db import models
from django.utils.translation import ugettext_lazy as _
from django.template.defaultfilters import slugify

from dcolumn.common.model_mixins import (
    UserModelMixin, TimeModelMixin, StatusModelMixin, StatusModelManagerMixin)

from .manager import dcolumn_manager

log = logging.getLogger('dcolumn.models')


#
# DynamicColumn
#
class DynamicColumnManager(StatusModelManagerMixin):

    def get_fk_slugs(self):
        result = {}

        for record in self.active():
            if record.value_type == self.model.CHOICE:
                name = dcolumn_manager.choice_relation_map.get(record.relation)
                result[name] = record.slug

        return result


class DynamicColumn(TimeModelMixin, UserModelMixin, StatusModelMixin):
    BOOLEAN = 1
    CHOICE = 2
    DATE = 3
    FLOAT = 4
    NUMBER = 5
    TEXT = 6
    TEXT_BLOCK = 7
    VALUE_TYPES = (
        (BOOLEAN, _("Boolean")),
        (CHOICE, _("Choice")),
        (DATE, _("Date")),
        (FLOAT, _("Floating Point")),
        (NUMBER, _("Number")),
        (TEXT, _("Text")),
        (TEXT_BLOCK, _("Text Block")),
        )
    VALUE_TYPES_MAP = dict(VALUE_TYPES)
    NO = False
    YES = True
    YES_NO = (
        (NO, _("No")),
        (YES, _("Yes"))
        )

    name = models.CharField(
        verbose_name=_("Name"), max_length=50,
        help_text=_("Enter a column name."))
    preferred_slug = models.SlugField(
        verbose_name=_("Preferred Slug"), null=True, blank=True,
        help_text=_(u"If you don't want the slug to change when the name "
                    u"changes enter a slug here. However, if you change this "
                    u"field the slug will track it."))
    slug = models.SlugField(
        verbose_name=_("Slug"), editable=False,
        help_text=_(u"This field is normally created from the name field, "
                    u"however, if you want to prevent it from changing when "
                    u"the name changes enter a preferred slug above."))
    value_type = models.IntegerField(
        verbose_name=_("Value Type"), choices=VALUE_TYPES,
        help_text=_("Choose the value type."))
    relation = models.IntegerField(
        verbose_name=_("Choice Relation"), null=True, blank=True,
        help_text=_("Choose the Choice Relation Type."))
    required = models.BooleanField(
        verbose_name=_("Required Field"), choices=YES_NO, default=NO,
        help_text=_("If this field is required based on business rules then "
                    "choose 'Yes'."))
    store_relation = models.BooleanField(
        verbose_name=_("Store Relation Value"), choices=YES_NO, default=NO,
        help_text=_("Store the literal value not the primary key of the "
                    "relation (used when choices change often). The most "
                    "common usage is the default 'No', to not store."))
    location = models.IntegerField(
        verbose_name=_("Display Location"),
        choices=dcolumn_manager.css_containers,
        help_text=_("Choose a display location."))
    order = models.PositiveSmallIntegerField(verbose_name='Display Order')

    objects = DynamicColumnManager()

    class Meta:
        ordering = ('location', 'order', 'name',)
        verbose_name = _("Dynamic Column")
        verbose_name_plural = _("Dynamic Columns")

    def save(self):
        if self.preferred_slug:
            self.slug = self.preferred_slug
        else:
            self.slug = slugify(self.name)

        super(DynamicColumn, self).save()

    def __unicode__(self):
        return self.name

    def _relation_producer(self):
        result = ''

        if self.relation is not None:
            result = dcolumn_manager.choice_relation_map.get(self.relation, '')

        return result
    _relation_producer.short_description = _("Relation")

    def _collection_producer(self):
        result = []

        for collection in self.column_collection.all():
            result.append('<span>{}</span>'.format(collection.name))

        return ", ".join(result)
    _collection_producer.short_description = _("Collections")
    _collection_producer.allow_tags = True

    def get_choice_relation_object_and_field(self):
        return dcolumn_manager.get_relation_model_field(self.value_type)



#
# ColumnCollection
#
class ColumnCollectionManager(StatusModelManagerMixin):

    def get_column_collection(self, name, unassigned=False):
        """
        Get the query set for the named collection. If unassigned is True add
        the unassigned dynamic colums to the query set.

        :Parameters:
            name : `str`
              Name of the column collection.
            unassigned : `bool`
              Also get items that are not assigned to a column collection yet.
        """
        log.debug("Collection name: %s, unassigned: %s", name, unassigned)
        queryset = self.none()

        try:
            queryset = self.active().get(name=name).dynamic_column.active()
        except self.model.DoesNotExist:
            pass # This is caught in the form.

        if unassigned:
            queryset = queryset | DynamicColumn.objects.active().filter(
                column_collection=None)

        return queryset

    def serialize_columns(self, name, obj=None, by_slug=False):
        records = self.get_column_collection(name)
        result = OrderedDict()

        if obj:
            key_value_map = obj.serialize_key_value_pairs()

        if by_slug:
            key = u'slug'
        else:
            key = u'pk'

        for record in records:
            rec = result.setdefault(getattr(record, key), {})
            rec[u'pk'] = record.pk
            rec[u'name'] = record.name
            rec[u'slug'] = record.slug
            rec[u'value_type'] = record.value_type

            if record.relation:
                rec[u'relation'] = record.relation
                rec[u'store_relation'] = record.store_relation

            rec[u'required'] = record.required
            # We convert the list to a dict because css_container_map may
            # not be keyed with integers.
            rec[u'location'] = dict(dcolumn_manager.css_containers).get(
                record.location, u'')
            rec[u'order'] = record.order
            if obj: rec[u'value'] = key_value_map.get(record.pk, u'')

        return result

    def get_active_relation_items(self, name):
        records = self.get_column_collection(name)
        return [dcolumn_manager.choice_relation_map.get(record.relation)
                for record in records if record.relation]

    def get_collection_choices(self, name, use_pk=False):
        """
        Returns a set of choices for a list of options on an HTML select tag.
        Normally the slug is returnd as the option value, however if `use_pk`
        is `True` then the value attribute will get the pk of the record.
        """
        records = self.get_column_collection(name)
        choices = [(use_pk and r.pk or r.slug, r.name) for r in records]
        return choices


class ColumnCollection(TimeModelMixin, UserModelMixin, StatusModelMixin):
    name = models.CharField(
        verbose_name=_("Name"), unique=True, max_length=50,
        help_text=_("Enter a unique name for this record."))
    dynamic_column = models.ManyToManyField(
        DynamicColumn, verbose_name=_("Dynamic Column"),
        related_name='column_collection')

    objects = ColumnCollectionManager()

    class Meta:
        verbose_name = _("Column Collection")
        verbose_name_plural = _("Column Collections")

    def __unicode__(self):
        return u"{}-{}".format(self.name, self.updated.isoformat())


#
# CollectionBase
#
class CollectionBaseManagerBase(models.Manager):

    def get_all_slugs(self):
        return [r.slug for r in DynamicColumn.objects.all().order_by('slug')]

    def get_all_fields(self):
        return [unicode(field)
                for field in self.model._meta.get_all_field_names()
                if u'collection' not in field and field != u'keyvalue_pairs']

    def get_all_fields_and_slugs(self):
        result = self.get_all_slugs() + self.get_all_fields()
        result.sort()
        return result


class CollectionBase(TimeModelMixin, UserModelMixin, StatusModelMixin):
    column_collection = models.ForeignKey(
        ColumnCollection, verbose_name=_("Column Collection"),
        help_text=_(u"Choose the version of the dynamic columns you want "
                    u"for all Collections."))

    def save(self, *args, **kwargs):
        log.debug("kwargs: %s", kwargs)
        super(CollectionBase, self).save(*args, **kwargs)

    def serialize_key_value_pairs(self):
        result = {}

        for kv in self.keyvalue_pairs.all():
            result[kv.dynamic_column_id] = kv.value

        return result

    def get_dynamic_column(self, slug):
        try:
            dc = self.column_collection.dynamic_column.get(slug=slug)
        except DynamicColumn.DoesNotExist as e:
            log.error("DynamicColumn with slug '%s' does not exist.", slug)
            # Cannot get a dynamic column if not in the collection already.
            dc = None

        return dc

    def get_key_value_pair(self, slug, field):
        dc = self.get_dynamic_column(slug)

        try:
            obj = self.keyvalue_pairs.get(dynamic_column=dc)

            model_meta = dc.get_choice_relation_object_and_field()

            if model_meta and not dc.store_relation:
                model, notused = model_meta
                value = model.objects.get_value_by_pk(obj.value, field=field)
            else:
                value = obj.value
        except self.keyvalue_pairs.model.DoesNotExist as e:
            log.error("Could not find value for slug '%s'.", slug)
            value = u''

        return value

    def set_key_value_pair(self, slug, value, field=None, force=False):
        """
        This method sets an arbitrary key/value pair, it will log an error
        if the key/value pair could not be found.

        If value contains the value 'increment' or 'decrement' the value
        associated with the slug will be incremented or decremented.

        Arguments:
          slug  -- The slug associated with the key value pair.
          value -- The value can be textor an object to get the value from.
          field -- The field used to get the value on the object.
          force -- Default is False, do not save empty strings or None objects
                   else True save empty strings only.
        """
        if (force and value == u'') or value not in (None, u''):
            dc = self.get_dynamic_column(slug)

            if dc:
                if dc.store_relation and field:
                    value = getattr(value, field)
                elif hasattr(value, field):
                    value = getattr(value, field)
                elif hasattr(value, u'pk'):
                    value = value.pk
                elif isinstance(value, (datetime.datetime, datetime.date)):
                    value = value.strftime(u"%Y-%m-%d")

                kv, created = self.keyvalue_pairs.get_or_create(
                    dynamic_column=dc, defaults={u'value': value})

                if not created:
                    if u'increment' == value and kv.value.isdigit():
                        value = int(kv.value) + 1
                    elif u'decrement' == value and kv.value.isdigit():
                        value = int(kv.value) - 1

                    kv.value = value
                    kv.save()


#
# KeyValue
#
class KeyValueManager(models.Manager):
    pass


class KeyValue(models.Model):
    collection = models.ForeignKey(
        CollectionBase, verbose_name=_("Collection Type"),
        related_name='keyvalue_pairs')
    dynamic_column = models.ForeignKey(
        DynamicColumn, verbose_name=_("Dynamic Column"),
        related_name='keyvalue_pairs')
    value = models.TextField(verbose_name=_("Value"), null=True, blank=True)

    objects = KeyValueManager()

    class Meta:
        ordering = ('dynamic_column__location', 'dynamic_column__order',)
        verbose_name = _("Key Value")
        verbose_name_plural = _("Key Values")

    def __unicode__(self):
        return self.value



#
# dcolumn/common/choice_mixins.py
#

import logging
import inspect

from django.utils.translation import ugettext_lazy as _

from dcolumn.common import ChoiceManagerImplementation

log = logging.getLogger('dcolumn.choices')


#
# InspectChoice
#
class InspectChoice(object):

    def __init__(self):
        self._path, self._caller_name = self._caller_info(skip=4)

    def _caller_info(self, skip=2):
        """
        Get a name of a caller in the format module.class.method.

        :Parameters:
          skip : `int`
            Specifies how many levels of stack to skip while getting caller
            name. skip=1 means 'who calls me', skip=2 'who calls my caller'
            etc.

        An empty string is returned if skipped levels exceed stack height

        See: https://gist.github.com/techtonik/2151727
        """
        stack = inspect.stack()
        start = 0 + skip

        if len(stack) < start + 1:
            return u''

        parentframe = stack[start][0]    
        name_list = []
        module = inspect.getmodule(parentframe)
        # `modname` can be None when frame is executed directly in console
        # TODO(techtonik): consider using __main__

        if module:
            name_list.append(module.__name__)

        # detect classname
        if u'self' in parentframe.f_locals:
            # I don't know any way to detect call from the object method
            # XXX: there seems to be no way to detect static method call--it
            # will be just a function call
            name_list.append(parentframe.f_locals[u'self'].__class__.__name__)

        codename = parentframe.f_code.co_name

        if codename != u'<module>':  # top level usually
            name_list.append( codename ) # function or a method

        del parentframe
        #log.debug("name_list: %s", name_list)
        return name_list

    @classmethod
    def set_model(self, method):
        """
        A decorator to set the choice/model object of the calling class.
        """
        def wrapper(this):
            modules = __import__(this._path, globals(), locals(),
                                 (this._caller_name,), -1)
            this.model = getattr(modules, this._caller_name)
            return method(this)

        # Make the wrapper look like the decorated method.
        wrapper.__name__ = method.__name__
        wrapper.__dict__ = method.__dict__
        wrapper.__doc__ = method.__doc__
        return wrapper


#
# BaseChoiceManager
#
class BaseChoiceManager(InspectChoice, ChoiceManagerImplementation):
    VALUES = None
    FIELD_LIST = None

    def __init__(self):
        super(BaseChoiceManager, self).__init__()
        self.containers = []
        self.container_map = {}

        if not self.VALUES:
            raise TypeError(_(u"Must set the '_VALUES' object to valid "
                              u"choices for the container object."))

        if not self.FIELD_LIST and len(self.FIELD_LIST) > 1:
            raise TypeError(_(u"Must provide fields to populate for your "
                              u"choices."))

    @InspectChoice.set_model
    def dynamic_column(self):
        if not self.containers:
            for pk, values in enumerate(self.VALUES, start=1):
                obj = self.model()
                setattr(obj, self.FIELD_LIST[0], pk)

                if isinstance(values, (tuple, list)):
                    for idx in xrange(len(values)):
                        setattr(obj, self.FIELD_LIST[idx+1], values[idx])
                else:
                    setattr(obj, self.FIELD_LIST[1], values)

                self.containers.append(obj)

            self.container_map.update(dict([(cont.pk, cont)
                                            for cont in self.containers]))

        return self.containers

    def get_value_by_pk(self, pk, field=None):
        self.dynamic_column()
        value = u''
        pk = int(pk)

        if pk != 0:
            try:
                obj = self.container_map.get(pk)
                value = getattr(obj, field)
            except AttributeError as e:
                log.error("Access to PK %s failed, %s", pk, e)

        return value

    def get_choices(self, field, comment=True):
        choices = [(obj.pk, getattr(obj, field))
                   for obj in self.dynamic_column()]

        if comment:
            choices.insert(
                0, (0, _("Please choose a {}".format(self.model.__name__))))

        return choices

    def get_choice_map(self, field):
        return dict([(getattr(obj, field), obj.pk)
                     for obj in self.dynamic_column()])



#
# dcolumn/common/decorators.py
#

from django.contrib.auth import REDIRECT_FIELD_NAME
from django.contrib.auth.decorators import login_required

from dcolumn.dcolumns.manager import dcolumn_manager


def dcolumn_login_required(function=None,
                           redirect_field_name=REDIRECT_FIELD_NAME,
                           login_url=None):
    """
    Decorator for views that checks that the user is logged in, redirecting
    to the log-in page if necessary.
    """
    if dcolumn_manager.get_api_auth_state():
        return function

    return login_required(function, redirect_field_name=redirect_field_name,
                          login_url=login_url)



#
# dcolumn/dcolumns/admin.py
#

from django.contrib import admin
from django.utils.translation import ugettext_lazy as _

from dcolumn.common.admin_mixins import UserAdminMixin

from .models import DynamicColumn, ColumnCollection, KeyValue
from .forms import DynamicColumnForm, ColumnCollectionForm, KeyValueForm


#
# KeyValue
#
class KeyValueInline(admin.TabularInline):
    fieldsets = (
        (None, {'fields': ('dynamic_column', 'value',)}),
        )
    ordering = ('dynamic_column__location', 'dynamic_column__order',)
    extra = 0
    model = KeyValue
    form = KeyValueForm

    class Media:
        js = ('dcolumn/js/jquery-1.11.1.min.js', 'dcolumn/js/jquery.cookie.js',
              'dcolumn/js/inheritance.js', 'dcolumn/js/dynamic-column.js',)


#
# ColumnCollection
#
class ColumnCollectionAdmin(UserAdminMixin):
    fieldsets = (
        (None, {'fields': ('name', 'dynamic_column',)}),
        (_('Status'), {'classes': ('collapse',),
                       'fields': ('active', 'creator', 'created', 'updater',
                                  'updated',)}),
        )
    readonly_fields = ('creator', 'created', 'updater', 'updated',)
    list_display = ('name', 'updater', 'updated',)
    filter_horizontal = ('dynamic_column',)
    form = ColumnCollectionForm

    class Media:
        js = ('dcolumn/js/jquery-1.11.1.min.js', 'dcolumn/js/inheritance.js',
              'dcolumn/js/column-collection.js',)
        css = {u'all': ('dcolumn/css/column-collection.css',)}


#
# DynamicColumn
#
class DynamicColumnAdmin(UserAdminMixin):
    fieldsets = (
        (None, {'fields': ('name', 'value_type', 'relation',
                           'store_relation',)}),
        (_('Screen Location'), {'fields': ('location', 'order', 'required',)}),
        (_('Status'), {'classes': ('collapse',),
                       'fields': ('preferred_slug', 'slug', 'active',
                                  'creator', 'created', 'updater',
                                  'updated',)}),
        )
    readonly_fields = ('creator', 'slug', 'updater', 'created', 'updated',)
    list_display = ('name', '_collection_producer', 'value_type',
                    '_relation_producer', 'store_relation', 'required',
                    'location', 'order', 'updated', 'active',)
    list_editable = ('location', 'order', 'active',)
    ordering = ('column_collection__name', 'location', 'order', 'name',)
    form = DynamicColumnForm


admin.site.register(ColumnCollection, ColumnCollectionAdmin)
admin.site.register(DynamicColumn, DynamicColumnAdmin)



#
# dcolumn/common/__init__.py
#
"""
Implementation class.

by: Carl J. Nobile

email: carl.nobile@gmail.com
"""
__docformat__ = "restructuredtext en"


#
# ChoiceManagerImplementation
#
class ChoiceManagerImplementation(object):
    """
    Manditory defined method.
    """

    def get_value_by_pk(self, pk, field=None):
        raise NotImplementedError("Must implement 'get_value_by_pk'.")

    def get_choices(self, field, comment=True):
        raise NotImplementedError("Must implement 'get_choices'.")

    def dynamic_column(self, active=True):
        raise NotImplementedError("Must implement 'dynamic_column'.")

    def get_choice_map(self, field, active=True):
        raise NotImplementedError("Must implement 'get_choice_map'.")



#
# dcolumn/dcolumns/views.py
#

import logging

from django.shortcuts import render
from django.views.generic import TemplateView
from django.core.urlresolvers import reverse
from django.utils.decorators import method_decorator

from dcolumn.common.view_mixins import JSONResponseMixin
from dcolumn.common.decorators import dcolumn_login_required
from .models import DynamicColumn, ColumnCollection
from .manager import dcolumn_manager

log = logging.getLogger('dcolumn.views')


class ContextDataMixin(object):

    def get_dynamic_column_context_data(self, class_name=u'', **kwargs):
        context = {}
        fk_slugs = DynamicColumn.objects.get_fk_slugs()
        name = dcolumn_manager.get_collection_name(class_name)

        for model_name in ColumnCollection.objects.get_active_relation_items(
            name):
            model, field = dcolumn_manager.choice_map.get(model_name)
            objects = context.setdefault(u'dynamicColumns', {})
            values = [(r.pk, getattr(r, field))
                      for r in model.objects.dynamic_column()]
            values.insert(0, (0, "Choose a value"))
            objects[fk_slugs.get(model_name)] = values
            log.debug("model_name: %s, model: %s, field: %s, fk_slugs: %s, "
                      "values: %s", model_name, model, field, fk_slugs, values)

        log.debug("context: %s", context)
        return context

    def get_relation_context_data(self, class_name=u'', obj=None, **kwargs):
        form = kwargs.get(u'form')

        if form:
            relations = form.get_display_data()
        else:
            name = dcolumn_manager.get_collection_name(class_name)
            relations = ColumnCollection.objects.serialize_columns(
                name, obj=obj)

        log.debug("relations: %s", relations)
        return {u'relations': relations}


#
# CollectionAJAXView
#
class CollectionAJAXView(JSONResponseMixin, TemplateView, ContextDataMixin):
    http_method_names = ('get',)

    @method_decorator(dcolumn_login_required)
    def dispatch(self, *args, **kwargs):
        return super(CollectionAJAXView, self).dispatch(*args, **kwargs)

    def get_context_data(self, **kwargs):
        """
        Get context data for the KeyValue objects.

        Do not call super on get_context_data, it puts self into the context
        which is not valid JSON.
        """
        log.debug("kwargs: %s", kwargs)
        context = {u'valid': True}

        try:
            context.update(self.get_dynamic_column_context_data(**kwargs))
            context.update(self.get_relation_context_data(**kwargs))
        except Exception, e:
            context[u'valid'] = False
            context[u'message'] = "Error occured: {}".format(e)
            log.error(context[u'message'], exc_info=True)

        return context

collection_ajax_view = CollectionAJAXView.as_view()


#
# CollectionCreateUpdateViewMixin
#
class CollectionCreateUpdateViewMixin(ContextDataMixin):

    def get_initial(self):
        return {'request': self.request}

    def get_context_data(self, **kwargs):
        """
        Get context data for the KeyValue objects.
        """
        kwargs[u'class_name'] = self.model.__name__
        context = super(CollectionCreateUpdateViewMixin,
                        self).get_context_data(**kwargs)
        context.update(self.get_dynamic_column_context_data(**kwargs))
        context.update(self.get_relation_context_data(**kwargs))
        context.update({u'css': dcolumn_manager.css_container_map})
        return context


#
# CollectionDetailViewMixin
#
class CollectionDetailViewMixin(ContextDataMixin):

    def get_context_data(self, **kwargs):
        """
        Get context data for the KeyValue objects.
        """
        kwargs[u'class_name'] = self.model.__name__
        context = super(CollectionDetailViewMixin,
                        self).get_context_data(**kwargs)
        context.update(self.get_dynamic_column_context_data(**kwargs))
        context.update(self.get_relation_context_data(
            obj=self.object, **kwargs))
        context.update({u'css': dcolumn_manager.css_container_map})
        return context



#
# dcolumn/dcolumns/urls.py
#

from django.conf.urls import *


urlpatterns = patterns(
    'dcolumn.dcolumns.views',
    url(r'api/collections/(?P<class_name>\w+)/$', 'collection_ajax_view',
        name="api-collections"),
    )



#
# dcolumn/dcolumns/manager.py
#

import logging

from django.conf import settings
from django.utils.translation import ugettext_lazy as _

log = logging.getLogger(u'dcolumn.manager')


class DynamicColumnManager(object):
    """
    This class manages the dynamic columns.
    """
    __shared_state = {}
    _relations = [(0, _("---------"))]
    _relation_map = None
    _relation_numbers = set()
    _choice_map = {}
    _css_containers = []
    _css_container_map = {}

    def __init__(self):
        self.__dict__ = self.__shared_state

    def register_choice(self, choice, relation_num, field):
        """
        Register choice field types. These can be Foreign Key or
        multiple-choice non-database columns. ManyToMany is not supported
        at this time.

        :Parameters:
          choice : ClassType
            This can be either a Django model or choice object. A choice object
            mimics a model class so that this manager can work with them
            as if they were Django models.

          relation_num : int
            A numeric identifier for the `choice` used as the HTML select
            option value.

          field : str
            A field from the model or choice object used as the HTML select
            option text.
        """
        if relation_num in self._relation_numbers:
            msg = (u"Invalid relation number {} is already used. [choice: {}, "
                   u"field: {}]").format(relation_num, choice, field)
            log.critical(msg)
            raise ValueError(msg)

        if (not hasattr(choice, 'objects') or
            not hasattr(choice.objects, 'dynamic_column')):
            msg = (u"Invalid 'choice' object '{}', must have a "
                   u"'dynamic_column' manager class method.").format(choice)
            log.critical(msg)
            raise AttributeError(msg)

        self._test_field(choice, field)
        self._relation_numbers.add(relation_num)
        self._relations.append((relation_num, choice.__name__))
        self._choice_map[choice.__name__] = (choice, field)
        log.debug("choice: %s, relation_num: %s, field: %s, relations: %s, "
                  "choice_map: %s", choice, relation_num, field,
                  self._relations, self._choice_map)

    def _test_field(self, choice, field):
        """
        Test that the specified field is a member object on the instantiated
        class object.

        :Parameters:
          choice : ClassType
            This can be either a Django model or choice object. A choice object
            mimics a model class so that this manager can work with them
            as if they were Django models.
          field : str
            A field from the model or choice object used as the HTML select
            option text.
        """
        obj = choice()

        if not hasattr(obj, field):
            msg = u"The '{}' object does not have the field '{}'".format(
                choice.__name__, field)
            log.critical(msg)
            raise AttributeError(msg)

    @property
    def choice_relations(self):
        """
        A property that returns the HTML select tag choices.

        :Returns:
          A list of the choices.
        """
        return self._relations

    @property
    def choice_relation_map(self):
        """
        A property that returns a dict (map) of the choices.

        :Returns:
          A dict of the choices.
        """
        if not self._relation_map:
            self._relation_map = dict(self._relations)

        return self._relation_map

    @property
    def choice_map(self):
        """
        A property that returns a dict (map). This property is used internally
        by `dcolumn`.

        :Returns:
          A dict where the key is the choice name (model or choice type) and
          the value is a tuple of the model/choice object and the field.
        """
        return self._choice_map

    def register_css_containers(self, container_list):
        """
        Register the CSS container objects. This method is usually called in
        the settings file.

        :Parameters:
          container_list : list or tuple
            A list of the CSS classes or ids that will determine the location
            on the page of the various dynamic columns.
        """
        if isinstance(container_list, (list, tuple)):
            if len(container_list) <= 0:
                msg = (u"Must supply at least one CSS container. The format "
                       u"can be in either of these formats: "
                       u"(u'container_top', u'container_bottom',) or "
                       u"((u'top', u'container_top'), "
                       u"(u'bottom', u'container_bottom'))")
                log.critical(msg)
                raise TypeError(msg)

            self._css_container_map.clear()

            if isinstance(container_list[0], (list, tuple)):
                self._css_container_map.update(dict(container_list))
                self._css_containers[:] = [
                    (key, css) for key, css in enumerate(
                        [css for name, css in container_list])]
            else:
                self._css_container_map.update(
                    {name: css for name, css in enumerate(container_list)})
                self._css_containers[:] = [
                    (key, css) for key, css in enumerate(container_list)]

    @property
    def css_containers(self):
        """
        A property that returns the CSS container classes or ids and is used
        internally as a choice to location in the DynamicColumn model.

        :Returns:
          A list of tuples where the tuple is (num, text)
        """
        return self._css_containers

    @property
    def css_container_map(self):
        """
        A property that returns a dict where the key is the CSS container
        number and the value is the CSS class or id. This property should be
        used in templates to desgnate location in the HTML.

        :Returns:
          A dict of the CSS containers.
        """
        return self._css_container_map

    def get_collection_name(self, model_name):
        """
        Gets the `ColumnCollection` instance name.

        :Parameters:
          model_name : str
            The key name used in the `settings.DYNAMIC_COLUMNS.COLLECTIONS`.

        :Returns:
          The `ColumnCollection` instance name.
        """
        model_name = model_name.lower()
        item_names = dict([
            (k.lower(), v)
            for k, v in settings.DYNAMIC_COLUMNS.get(u'COLLECTIONS').items()])

        if model_name not in item_names:
            msg = _("Invalid model name: {}".format(model_name))
            raise KeyError(msg)

        return item_names.get(model_name, u'')

    def get_api_auth_state(self):
        """
        Gets the value of settings.DYNAMIC_COLUMNS.INACTIVATE_API_AUTH.
        """
        return settings.DYNAMIC_COLUMNS.get(u'INACTIVATE_API_AUTH', False)

    def get_relation_model_field(self, relation):
        """
        Gets the field used in the HTML select option text value.

        :Parameters:
          relation : int
            The value in the `DynamicColumn` relation field.

        :Returns:
          The field used in the HTML select option text value.
        """
        return self.choice_map.get(self.choice_relation_map.get(relation))

dcolumn_manager = DynamicColumnManager()



#
# dcolumn/common/admin_mixins.py
#

"""
Mixins used in the Django admin.

by: Carl J. Nobile

email: carl.nobile@gmail.com
"""
__docformat__ = "restructuredtext en"

import datetime

from django.contrib import admin


class UserAdminMixin(admin.ModelAdmin):
    """
    Admin mixin that should be used in any model implimented with dynamic
    columns.
    """
    def save_model(self, request, obj, form, change):
        """
        When saving a record from the admin the `creator` should be updated
        with the request user object if `change` is `False`. The `updater` is
        always updated withthe request user object.

        :Parameters:
          request : HttpRequest
            Django request object.
          obj : Model
            Django model object
          form : Form
            Django form object.
          change : bool
            If `True` the record was updated, if `False` the record was created.
        """
        if change is False:
            obj.creator = request.user

        obj.updater = request.user
        super(UserAdminMixin, self).save_model(request, obj, form, change)



import os
from setuptools import setup

with open(os.path.join(os.path.dirname(__file__), 'README.md')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-dcolumns',
    version='0.2.0',
    packages=['dcolumn.dcolumns', 'dcolumn.common',],
    include_package_data=True,
    license='MIT',
    description=('An app to give any Django database model the ability to '
                 'dynamically add fields.'),
    long_description=README,
    url='https://github.com/cnobile2012/dcolumn',
    author='Carl J. Nobile',
    author_email='carl.nobile@gmail.com',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        ],
    )



#
# dcolumn/dcolumns/templatetags/autodisplay.py
#

"""
Template tags used for displaying dynamic columns.

by: Carl J. Nobile

email: carl.nobile@gmail.com
"""
__docformat__ = "restructuredtext en"

import os, logging
import datetime
from StringIO import StringIO

from django import template
from django.utils.safestring import mark_safe

from dcolumn.dcolumns.models import DynamicColumn, KeyValue
from dcolumn.dcolumns.manager import dcolumn_manager

log = logging.getLogger(u'dcolumn.templates')
register = template.Library()


#
# auto_display
#
# NOTE: Formatting of the doc string is to conform with django docs not epydoc.
#
@register.tag(name='auto_display')
def auto_display(parser, token):
    """
    This tag returns an HTML element.

    Arguments::

      relation -- One object from the 'relations' template object.
      prefix   -- A keyword argument used as a prefix in the element id and
                  name attributes.
      option   -- A keyword argument that is in the 'dynamicColumns' context.
                  The entire 'dynamicColumns' context can be supplied or just
                  the object for this relation.
      display  -- True use <span> for all tags else False use default tag types.

    Assume data structures below for examples::

      {
        'dynamicColumns': {
          'book': [
            [0, 'Choose a value'],
            [2, 'HTML5 Pocket Reference'],
            [1, 'SQL Pocket Guide'],
            [3, 'Raspberry Pi Hacks']
          ]
        },
        'relations': {
          '1': {
            'name': 'Project ID',
            'required': false,
            'relation': null,
            'location': 'top-container',
            'value_type': 0,
            'order': 3,
            'value': '12345'
          },
          '12': {
            'name': 'Book',
            'required': false,
            'relation': 1,
            'location': 'bottom-container',
            'key': 'book',
            'value_type': 6,
            'order': 1,
            'value': 2
          }
        }
      }

    Usage Examples::

      {% auto_display relation %}

      {% auto_display relation prefix=test- %}

      {% auto_display relation options=books %}

      {% auto_display relation options=books display=True %}

      {% auto_display relation prefix=test- options=dynamicColumns %}
    """
    tokens = token.split_contents()
    size = len(tokens)
    keywords = (u'prefix', u'options', u'display')
    kwargs = {u'prefix': u'', u'options': None, u'display': u'False'}

    if size == 2:
        tag_name, relation = tokens
        kwargs = {}
    elif size == 3:
        tag_name, relation, value1 = tokens
        kwargs.update({k: v for k,d,v in [v.partition('=')
                                          for v in (value1,)]})
    elif size == 4:
        tag_name, relation, value1, value2 = tokens
        kwargs.update({k: v for k,d,v in [v.partition('=')
                                          for v in (value1, value2)]})
    elif size == 5:
        tag_name, relation, value1, value2, value3 = tokens
        kwargs.update({k: v for k,d,v in [v.partition('=')
                                          for v in (value1, value2, value3)]})
    else:
        msg = (u"Invalue number of arguments should be 1 - 4, "
               u"found: {}").format(size-1)
        raise template.TemplateSyntaxError(msg)

    if size > 2 and not all([key in keywords for key in kwargs]):
        msg = u"Invalid keyword name, should be one of {}".format(keywords)
        raise template.TemplateSyntaxError(msg)

    return AutoDisplayNode(tag_name, relation, **kwargs)


class AutoDisplayNode(template.Node):
    """
    Node class for the `auto_display` tag.
    """
    ERROR_MSG = (u"Invalid relation object--please note steps that "
                 u"led to this error and file a bug report.")
    YES_NO = ((1, u'Unknown'), (2, u'Yes'), (3, u'No'),)
    DISPLAY_TAG = u'<span id="{}">{}</span>'
    STORE_WRAPPER = u'<div class="storage-wrapper">{}{}</div>'
    ELEMENT_TYPES = {
        DynamicColumn.NUMBER: (u'<input id="{}" name="{}" type="number" '
                               u'value="{}" />'),
        DynamicColumn.TEXT: (u'<input id="{}" name="{}" size="50" type="text" '
                             u'value="{}" />'),
        DynamicColumn.TEXT_BLOCK: (u'<textarea class="large-text-field" '
                                   u'id="{}" name="{}" cols="40" rows="10"'
                                   u'>{}</textarea>\n'),
        DynamicColumn.DATE: (u'<input id="{}" type="date" name="{}" '
                             u'value="{}" />'),
        DynamicColumn.BOOLEAN: u'<select id="{}" name="{}">\n',
        DynamicColumn.FLOAT: (u'<input id="{}" name="{}" type="text" '
                              u'value="{}" />'),
        DynamicColumn.CHOICE: u'<select id="{}" name="{}">\n',
        }

    def __init__(self, tag_name, relation, prefix=u'', options=None,
                 display='False'):
        self.tag_name = tag_name
        self.relation = template.Variable(relation)
        self.prefix = prefix
        self.fk_options = options and template.Variable(options) or None
        self.display = eval(display)

    def render(self, context):
        """
        Render the results in an HTML friendly way.

        :Parameters:
          context
            The context as profided bt django.

        :Return:
          The HTML element rendered for a specific dynamic column slug.
        """
        try:
            relation = self.relation.resolve(context)
        except template.VariableDoesNotExist:
            relation = {}

        log.debug("relation: %s, display: %s", relation, self.display)

        if relation:
            value_type = relation.get(u'value_type')

            if self.display:
                elem = self.DISPLAY_TAG
            else:
                elem = self.ELEMENT_TYPES.get(value_type, u'')

            attr = u"{}{}".format(self.prefix, relation.get(u'slug'))

            # Choices are a special case since we need to determine
            # what the options will be.
            if value_type == DynamicColumn.CHOICE:
                try:
                    fk_options = self.fk_options.resolve(context)
                except Exception:
                    fk_options = {}

                options = self._find_options(relation, fk_options)

                if self.display:
                    elem = self._find_value(elem, attr, options, relation)
                else:
                    elem = self._add_options(elem, attr, options, relation)

                    if (relation.get(u'store_relation', False) and
                        u'selected' not in elem):
                        tmp_elem = self._find_value(
                            self.DISPLAY_TAG, u'store-' + attr, {}, relation)
                        elem = self.STORE_WRAPPER.format(elem, tmp_elem)
            elif value_type == DynamicColumn.BOOLEAN:
                if self.display:
                    elem = self._find_value(elem, attr, self.YES_NO, relation)
                else:
                    elem = self._add_options(elem, attr, self.YES_NO, relation)
            elif self.display:
                elem = elem.format(u"id-" + attr, relation.get(u'value', u''))
            else:
                elem = elem.format(u"id-" + attr, attr,
                                   relation.get(u'value', u''))
        else:
            elem = u"<span>{}</span>".format(self.ERROR_MSG)

        return mark_safe(elem)

    def _find_options(self, relation, fk_options):
        """
        Find the options for this relation.

        The fk_option argument can either be the options we need in a list or
        tuple or a dict of all options for all choice type objects.

        :Parameters:
          relation : dict
            The meta data for a dynamic column.
          fk_options : list or dict
            Can be the list of a specific slug's display options or the full
            `dynamicColumns` dict object.

        :Returns:
          The list of options.
        """
        if isinstance(fk_options, (list, tuple,)):
            options = fk_options
        else:
            slug = relation.get(u'slug')

            if slug:
                options = fk_options.get(slug, {})
            else:
                msg = u'Invalid key for relation, {}'.format(relation)
                log.error(msg)
                raise template.TemplateSyntaxError(msg)

        return options

    def _add_options(self, elem, attr, options, relation):
        """
        Find the value in the options if it exists and set the selected
        property on the appropriate option. In the case of a stored relation
        find the selected option by the actual value as the PK is not
        available in this case.

        :Parameters:
          elem : str
            The HTML element.
          attr : str
            Text object to be used when creating the element's attributes.
          options : list
            The options used when creating the HTML select element.
          relation : dict
            The meta data for a dynamic column.

        :Returns:
          The populated HTML element.
        """
        elem = elem.format("id-" + attr, attr)
        buff = StringIO(elem)
        buff.seek(0, os.SEEK_END)
        value = relation.get(u'value', u'')

        # Get the ID if the value is stored and not a pk.
        if relation.get(u'store_relation', False):
            value = [k for k, v in options if v == value]
            value = len(value) >= 1 and value[0] or 0

        value = unicode(value).isdigit() and int(value) or value

        for k, v in options:
            s = u""

            if k == value and unicode(value).isdigit() and int(value) != 0:
                s = u" selected"

            buff.write(u'<option value="{}"{}>{}</option>\n'.format(k, s, v))

        buff.write(u'</select>\n')
        elem = buff.getvalue()
        buff.close()
        return elem

    def _find_value(self, elem, attr, options, relation):
        """
        Produce the element filled with the value. This method is used only
        for display mode and the element is usually a <span>.

        :Parameters:
          elem : str
            The HTML element.
          attr : str
            Text object to be used when creating the element's attributes.
          options : list
            The options used when creating the HTML select element.
          relation : dict
            The meta data for a dynamic column.

        :Returns:
          The populated HTML element.
        """
        log.debug("elem: %s, attr: %s, options: %s, relation: %s",
                  elem, attr, options, relation)
        value = relation.get(u'value', u'')

        # Get the value is the ID then get the value.
        if relation.get(u'store_relation', False):
            if value == u'0':
                value = u''
        else:
            key = value.isdigit() and int(value) or value
            value = dict(options).get(key, u'')

        elem = elem.format("id-" + attr, value)
        return elem


#
# single_display
#
# NOTE: Formatting of the doc string is to conform with django docs not epydoc.
#
@register.tag(name='single_display')
def single_display(parser, token):
    """
    Returns a context variable containing a single value where the data type is
    based on the type of object. You will need to warp the result in an html
    tag of your choice.

    Arguments::

      obj  -- A CollectionBase derived model object.
      slug -- The slug for the DynamicColumn type.
      as   -- Manditory delimiter.
      name -- Name for context variable similar to the slug name.

    Usage Examples::

      {% single_display <object> <slug> as <context name>  %}

      {% single_display obj country as country %}
    """
    try:
        tag_name, obj, slug, delimiter, name = token.split_contents()
    except ValueError:
        msg = "{} tag requires four arguments."
        raise template.TemplateSyntaxError(
            msg.format(token.contents.split()[0]))

    if delimiter != u'as':
        raise template.TemplateSyntaxError(
            "The second argument must be the word 'as' found '{}'.".format(
                delimiter))

    return SingleDisplayNode(tag_name, obj, slug, name)


class SingleDisplayNode(template.Node):
    """
    Node class for the `single_display` tag.
    """

    def __init__(self, tag_name, obj, slug, name):
        self.tag_name = tag_name
        self.obj = template.Variable(obj)
        self.slug = slug
        self.name = name

    def _fix_boolean(self, dc, value):
        return {0: u'Unknown', 1: u'Yes', 2: u'No'}.get(value, u'')

    def _fix_choice(self, dc, value):
        choice_name = dcolumn_manager.choice_relation_map[dc.relation]
        obj, field = dcolumn_manager.choice_map[choice_name]
        value = int(value)

        for choice in obj.objects.dynamic_column():
            if getattr(choice, u'pk') == value:
                value = getattr(choice, field)
                break

        return (value != 0) and value or u''

    def _fix_date(self, dc, value):
        return datetime.date(*[int(v) for v in value.split('-') if v.isdigit()])

    def _fix_float(self, dc, value):
        return float(value)

    def _fix_number(self, dc, value):
        return value

    def _fix_text(self, dc, value):
        return value

    def _fix_text_block(self, dc, value):
        return value

    METHOD_MAP = {
        DynamicColumn.BOOLEAN: _fix_boolean,
        DynamicColumn.CHOICE: _fix_choice,
        DynamicColumn.DATE: _fix_date,
        DynamicColumn.FLOAT: _fix_float,
        DynamicColumn.NUMBER: _fix_number,
        DynamicColumn.TEXT: _fix_text,
        DynamicColumn.TEXT_BLOCK: _fix_text_block,
        }

    def render(self, context):
        try:
            obj = self.obj.resolve(context)
        except template.VariableDoesNotExist:
            obj = None

        value = u''

        try:
            key_value = obj.keyvalue_pairs.get(dynamic_column__slug=self.slug)
            dc = key_value.dynamic_column
            value_type = dc.value_type

            if key_value.value:
                if dc.store_relation and not key_value.value.isdigit():
                    value_type = DynamicColumn.TEXT

                value = self.METHOD_MAP[value_type](self, dc, key_value.value)
        except KeyValue.DoesNotExist:
            log.warn("KeyValue pair does not exist for slug %s", self.slug)

        context[self.name] = value
        return u''


#
# combine_contexts
#
# NOTE: Formatting of the doc string is to conform with django docs not epydoc.
#
@register.tag(name='combine_contexts')
def combine_contexts(parser, token):
    """
    Combines two context variables. Helps make the HTML a less cluttered.

    Arguments::

      obj      -- An object in the contect that has member objects.
      variable -- A variable that defines a member object name.

    Usage Examples::

      {% combine_contexts form.errors relation.slug %}
    """
    try:
        tag_name, obj, variable = token.split_contents()
    except ValueError:
        msg = "{} tag requires two arguments."
        raise template.TemplateSyntaxError(
            msg.format(token.contents.split()[0]))

    return CombineContextsNode(tag_name, obj, variable)


class CombineContextsNode(template.Node):
    """
    Node class for the `combine_contexts` tag.
    """

    def __init__(self, tag_name, obj, variable):
        self.tag_name = tag_name
        self.obj = template.Variable(obj)
        self.variable = template.Variable(variable)

    def render(self, context):
        result = self.obj.resolve(context).get(self.variable.resolve(context))
        return result and result or u''


if __name__ == '__main__':
    import sys, traceback
    from django.template import Template, Context

    context = {}
    context.update({u'relation': {u'name': u"Project Name"}})

    fk_options = {
        u"dynamicColumns": {
            u"book": [
                [0, u"Choose a value"],
                [2, u"HTML5 Pocket Reference"],
                [1, u"SQL Pocket Guide"],
                [3, u"Raspberry Pi Hacks"]
                ]
            },
        }

    for i in range(len(DynamicColumn.VALUE_TYPES)):
        relation = context.get(u'relation')
        relation[u'value_type'] = i

        if i == DynamicColumn.CHOICE:
            rel = context.get(u'relation')
            rel[u'key'] = u'book'
            cmd = (u"{% auto_display relation prefix=test- "
                   u"options=dynamicColumns %}")
            context.update(fk_options)
        else:
            cmd = u"{% auto_display relation %}"
            rel = context.get(u'relation')
            rel.pop(u'key', None)
            context.pop(u'dynamicColumns', None)

        try:
            print 'Test {}--{}'.format(i+1, cmd)
            c = Context(context)
            command = u"Element: {}".format(cmd)
            t = Template(u"{% load autodisplay %}" + command)
            print 'Context Data:', c
            print t.render(c)
            print
        except Exception, e:
            print "{}: {}\n".format(sys.exc_info()[0], sys.exc_info()[1])
            traceback.print_tb(sys.exc_info()[2])



#
# dcolumn/dcolumns/forms.py
#

import logging
import datetime

from django import forms
from django.utils.translation import ugettext_lazy as _

from .manager import dcolumn_manager
from .models import DynamicColumn, ColumnCollection, KeyValue

log = logging.getLogger('dcolumn.views')


#
# ColumnCollection
#
class ColumnCollectionForm(forms.ModelForm):

    def __init__(self, *args, **kwargs):
        super(ColumnCollectionForm, self).__init__(*args, **kwargs)
        log.debug("args: %s, kwargs: %s", args, kwargs)
        self.columns = ColumnCollection.objects.get_column_collection(
            self.instance.name, unassigned=True)
        self.fields[u'dynamic_column'].queryset = self.columns

    class Meta:
        model = ColumnCollection
        exclude = []

    def clean(self):
        if not self.columns:
            msg = ("No objects in the database, please create initial objects "
                   "in the Dynamic Columns model to be used for this "
                   "collection type.")
            log.error(msg)
            raise forms.ValidationError({u'dynamic_column': [msg]})

        return self.cleaned_data


#
# DynamicColumn
#
class DynamicColumnForm(forms.ModelForm):

    def __init__(self, *args, **kwargs):
        super(DynamicColumnForm, self).__init__(*args, **kwargs)
        self.fields[u'relation'] = forms.ChoiceField(
            widget=forms.Select, choices=dcolumn_manager.choice_relations)
        self.fields[u'relation'].required = False

    class Meta:
        model = DynamicColumn
        exclude = []

    def clean(self):
        cleaned_data = super(DynamicColumnForm, self).clean()
        value_type = cleaned_data.get(u'value_type')
        relation = cleaned_data.get(u'relation', u'0')
        relation = int(relation)
        log.debug("value_type: %s, relation: %s", value_type, relation)

        if value_type == DynamicColumn.CHOICE:
            if not relation:
                self._errors[u'relation'] = self.error_class(
                        [_(u"If the Value Type is a Choice then a relation "
                           u"must be entered.")])
        else:
            cleaned_data[u'relation'] = None

        return cleaned_data


#
# CollectionFormMixin
#
class CollectionFormMixin(forms.ModelForm):
    SPECIAL_CASE_MAP = {DynamicColumn.BOOLEAN: u'1',
                        DynamicColumn.CHOICE: u'0'}
    MAX_LENGTH_MAP = {DynamicColumn.NUMBER: 20,
                      DynamicColumn.TEXT: 250,
                      DynamicColumn.TEXT_BLOCK: 2000,
                      DynamicColumn.DATE: 20,
                      DynamicColumn.BOOLEAN: 1,
                      DynamicColumn.FLOAT: 305,
                      DynamicColumn.CHOICE: 12}

    class Meta:
        exclude = ['creator', 'updater', 'created', 'updated']

    def __init__(self, *args, **kwargs):
        super(CollectionFormMixin, self).__init__(*args, **kwargs)
        self.coll_name = dcolumn_manager.get_collection_name(
            self.Meta.model.__name__)
        self.relations = ColumnCollection.objects.serialize_columns(
            self.coll_name)
        self.fields[u'column_collection'].required = False

        if u'created' in self.fields:
            self.fields[u'created'].required = False

        if u'updated' in self.fields:
            self.fields[u'updated'].required = False

        log.debug("args: %s, kwargs: %s", args, kwargs)
        log.debug("fields: %s, data: %s", self.fields, self.data)

    def get_display_data(self):
        if self.instance and self.instance.pk is not None:
            for pk, value in self.instance.serialize_key_value_pairs().items():
                log.debug("pk: %s, value: %s", pk, value)
                relation = self.relations.setdefault(pk, {})
                # We only want to add new data not overwrite data that is
                # already there.
                if u'value' not in relation:
                    relation[u'value'] = value

        return self.relations

    def clean_column_collection(self):
        return ColumnCollection.objects.active().get(name=self.coll_name)

    def clean(self):
        cleaned_data = super(CollectionFormMixin, self).clean()
        self.validate_dynamic_fields()
        log.debug("cleaned_data: %s", cleaned_data)
        return cleaned_data

    def validate_dynamic_fields(self):
        for relation in  self.relations.values():
            log.debug("relation: %s", relation)

            for key, value in self.data.items():
                if key == relation.get(u'slug'):
                    value = self.validate_store_relation(relation, value)
                    value = self.validate_date_types(relation, key, value)
                    relation[u'value'] = value
                    self.validate_required(relation, key, value)
                    self.validate_value_type(relation, key, value)
                    self.validate_value_length(relation, key, value)

        log.warn("form.errors: %s", self._errors)

    def validate_store_relation(self, relation, value):
        """
        If 'store_relation' is False then return the value as is, should be a
        PK. If 'store_relation' is True then lookup in the choices using the
        PK and return the actual value.
        """
        if relation.get(u'store_relation', False):
            log.debug("value: %s, relation: %s", value, relation)
            data = dcolumn_manager.get_relation_model_field(
                relation.get('relation', u''))

            if len(data) == 2:
                model, field = data
                old_value = value
                value = value.isdigit() and int(value) or 0
                value = [getattr(r, field)
                         for r in model.objects.dynamic_column()
                         if value == r.pk]
                value = len(value) >= 1 and value[0] or old_value
                log.debug("value: %s, field: %s, model: %s",
                          value, field, model)

        return value

    # SHORT TERM FIX will only work with one data format type.
    _MONTHS = {u'january': 1, u'february': 2, u'march': 3, u'april': 4,
               u'may': 5, u'june': 6, u'july': 7, u'august': 8,
               u'september': 9, u'october': 10, u'november': 11,
               u'december': 12}

    def validate_date_types(self, relation, key, value):
        if relation.get(u'value_type') == DynamicColumn.DATE:
            if value:
                year = month = day = 0
                day_month, delim, year = value.partition(', ')

                if delim:
                    day, delim, month = day_month.partition(' ')
                    month = self._MONTHS.get(month.lower(), 0)
                else:
                    tmp = value.split(u'-')

                    if len(tmp) == 3:
                        year, month, day = tmp

                log.debug("year: %s, month: %s, day: %s", year, month, day)

                if not year.isdigit() or not month or not day.isdigit():
                    self._errors[key] = self.error_class(
                        [u"Invalid year, month, or day, found: {}".format(
                            value)])
                else:
                    value = datetime.date(int(year), int(month),
                                          int(day)).strftime(u'%Y-%m-%d')

        return value

    def validate_required(self, relation, key, value):
        if relation.get(u'required', False):
            value_type = relation.get(u'value_type')

            if (not value or value_type in self.SPECIAL_CASE_MAP and
                self.SPECIAL_CASE_MAP.get(value_type) == value):

                self._errors[key] = self.error_class(
                    [u"{} field is required.".format(relation.get(u'name'))])

    def validate_value_type(self, relation, key, value):
        value_type = relation.get(u'value_type')

        if value_type == DynamicColumn.NUMBER:
            if value and not value.isdigit():
                self._errors[key] = self.error_class(
                    [u"{} field is not a number.".format(
                        relation.get(u'name'))])

    def validate_value_length(self, relation, key, value):
        value_type = relation.get(u'value_type')

        # If store_relation is True then the storage type is DynamicColumn.TEXT.
        if relation.get(u'store_relation', False):
            value_type = DynamicColumn.TEXT

        log.debug("key: %s, value: %s", key, value)

        if len(value) > self.MAX_LENGTH_MAP.get(value_type):
                self._errors[key] = self.error_class(
                    [u"{} field is too long.".format(relation.get(u'name'))])

    def save(self, commit=True):
        inst = super(CollectionFormMixin, self).save(commit=False)
        request = self.initial.get('request')
        log.debug("request: %s, inst: %s, instance: %s",
                  request, inst, self.instance)

        if request:
            inst.updater = request.user
            inst.active = True

            if not hasattr(inst, u'creator') or not inst.creator:
                inst.creator = request.user

        if commit:
            inst.save()
            #self.save_m2m() # If we have any.
            self._save_keyvalue_pairs()

        return inst

    def _save_keyvalue_pairs(self):
        for pk, relation in self.get_display_data().items():
            required = relation.get(u'required', False)
            value = relation.get(u'value', u'')
            log.debug("pk: %s, slug: %s, value: %s",
                      pk, relation.get(u'slug'), relation.get(u'value'))

            try:
                obj, created = self.instance.keyvalue_pairs.get_or_create(
                    collection=self.instance, dynamic_column_id=int(pk),
                    defaults={u'value': value})
            except self.instance.MultipleObjectsReturned, e:
                log.error("Multiple records found for parent: %s, "
                          "dynamic_column_id: %s", self.instance,
                          dynamic_column_id=cc_id)
                raise e

            if not created:
                obj.value = value
                obj.save()


#
# KeyValue
#
class KeyValueForm(forms.ModelForm):

    def __init__(self, *args, **kwargs):
        super(KeyValueForm, self).__init__(*args, **kwargs)
        self.fields[u'value'].widget = forms.TextInput(
            attrs={u'size': 50, u'maxlength': 2000})
        log.debug("args: %s, kwargs: %s", args, kwargs)
        #log.debug("dir(self): %s", dir(self))

        if hasattr(self.instance, 'collection'):
            coll_name = self.instance.collection.column_collection.name
            columns = ColumnCollection.objects.get_column_collection(coll_name)
            self.fields[u'dynamic_column'].queryset = columns
        else:
            log.debug("instance: %s", self.instance)

    class Meta:
        model = KeyValue
        exclude = []
